import os
import unittest
#from common import DB_CREDENTIALS

@unittest.skip("Ignore untill we have environment ready")
class TestPresto(unittest.TestCase):
    def test_presto_ds(self):
        from mindsdb_datasources import PrestoDS

        # Create the datasource
        presto_ds = PrestoDS(
            query='SELECT * FROM HEALTHCARE_COSTS',
            host=DB_CREDENTIALS['presto']['host'],
            user=DB_CREDENTIALS['presto']['user'],
            password=DB_CREDENTIALS['presto']['password'],
            database=DB_CREDENTIALS['presto']['database'],
            schema=DB_CREDENTIALS['presto']['schema'],
            protocol=DB_CREDENTIALS['presto']['protocol'],
            port=DB_CREDENTIALS['presto']['port'],
            catalog=DB_CREDENTIALS['presto']['catalog'],
        )

        assert len(presto_ds.df) == 1338
        assert len(presto_ds.df.columns) == 7



import unittest
#from common import break_dataset

@unittest.skip("Ignore untill we have environment ready")
class TestBigQuery(unittest.TestCase):
    def test_bigquery_ds(self):
        from mindsdb_datasources import BigQueryDS

        LIMIT = 100

        query = """
            SELECT corpus AS title, COUNT(word) AS unique_words
            FROM `bigquery-public-data.samples.shakespeare`
            GROUP BY title
            ORDER BY unique_words
            DESC LIMIT {}
        """.format(LIMIT)

        bigquery_ds = BigQueryDS(
            query=query
        )

        bigquery_ds.df = break_dataset(bigquery_ds.df)

        assert len(bigquery_ds) == LIMIT




import pandas as pd
from crate import client

from mindsdb_datasources.datasources.data_source import SQLDataSource 


class CrateDS(SQLDataSource):
    """
    This method takes in the params required for the
    CrateDb connection
    params:
    user: str
    password: str
    hostname: list || str
    port: int
    verify_ssl : bool
    ca_cert: str
    schema: doc

    NOTE: This connection supports multiple host connections
    and SSL
    """

    def __init__(self, query, user, password, hostname, port=4201,
                 verify_ssl=False, ca_cert=None, schema="doc",
                 error_trace=True):
        super().__init__(query)
        self.user = user
        self.password = password
        self.hostname = hostname
        self.port = port
        self.verify_ssl = verify_ssl
        self.ca_cert = ca_cert
        self.schema = schema
        self.error_trace = error_trace

    def query(self, q):
        if self.verify_ssl:
            connection = client.connect(self.hostname,
                                        ca_cert=self.ca_cert,
                                        error_trace=self.error_trace,
                                        username=self.user,
                                        password=self.password,
                                        schema=self.schema)
        connection = client.connect(self.hostname,
                                    verify_ssl_cert=self.verify_ssl,
                                    error_trace=self.error_trace,
                                    username=self.user,
                                    password=self.password,
                                    schema=self.schema)
        df = pd.read_sql(q, con=connection)

        return df, self._make_colmap(df)

    def name(self):
        return 'CrateDB - {}'.format(self._query)




import pandas as pd
import requests

from mindsdb_datasources.datasources.data_source import SQLDataSource


class ClickhouseDS(SQLDataSource):
    def __init__(self, query, host='localhost', user='default', password=None,
                 port=8123, protocol='http'):

        if ' format ' in query.lower():
            raise Exception('Please refrain from adding a "FORMAT" statement to the query')

        super().__init__(query)

        self.host = host
        self.user = user
        self.password = password
        self.port = int(port)
        self.protocol = protocol

        if protocol not in ('https', 'http'):
            raise ValueError('Unexpected protocol {}'.fomat(protocol))

    def query(self, q):
        q = '{} FORMAT JSON'.format(q.rstrip(" ;\n"))
        params = {'user': self.user}
        if self.password is not None:
            params['password'] = self.password

        response = requests.post(
            f'{self.protocol}://{self.host}:{self.port}',
            data=q,
            params=params
        )

        try:
            data = response.json()['data']
        except Exception:
            raise Exception(f'Got an invalid response from the database: {response.text}')

        df = pd.DataFrame(data)

        return df, self._make_colmap(df)

    def name(self):
        return 'Clickhouse - {}'.format(self._query)



import os

import pandas as pd
import cx_Oracle

from mindsdb_datasources.datasources.data_source import SQLDataSource


class OracleDS(SQLDataSource):
    def __init__(self, query, sid='oracle', host='',
                 port, username='', password=''):
        super().__init__(query)
        self.host = host
        self.port = int(port)
        self.username = username
        self.password = password
        self.sid=sid

    def query(self, q):
        dsn_tns=cx_Oracle.makedsn(self.host,self.port,self.sid)
        con = cx_Oracle.connect(
            username=self.username,
            password=self.password,
            dsn_tns
        )

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Oracle - {}'.format(self._query)



import re

import certifi
import pandas as pd
from pymongo import MongoClient
from pandas.api.types import is_numeric_dtype

from mindsdb_datasources.datasources.data_source import DataSource

class MongoDS(DataSource):
    def __init__(self, query, collection, database='database',
                 host='127.0.0.1', port=None, user=None, password=None,
                 limit=None, sort_by=None):

        """params:
            limit: limit of requested documents
            sort_by: sort settings. it is a dict with fields as keys
                and '1'(ascending) '-1'(descending) as values
                example sort_by={'id': 1, 'vendor_id': -1}"""
        super().__init__()

        if not isinstance(query, dict):
            raise TypeError('query must be a dict')
        self._query = query

        if not re.match(r'^mongodb(\+srv)?:\/\/', host.lower()):
            port = int(port or 27017)

        self.collection = collection
        self.database = database
        self.limit = limit
        self.sort_by = sort_by
        self.host = host
        self.port = int(port or 27017)
        self.user = user
        self.password = password

    def _get_connection(self):
        kwargs = {
            'appname': 'MindsDB'
        }

        if isinstance(self.user, str) and len(self.user) > 0:
            kwargs['username'] = self.user

        if isinstance(self.password, str) and len(self.password) > 0:
            kwargs['password'] = self.password

        if re.match(r'\/\?.*tls=true', self.host.lower()):
            kwargs['tls'] = True

        if re.match(r'\/\?.*tls=false', self.host.lower()):
            kwargs['tls'] = False

        if re.match(r'.*\.mongodb.net', self.host.lower()):
            kwargs['tlsCAFile'] = certifi.where()
            if kwargs.get('tls', None) is None:
                kwargs['tls'] = True

        return MongoClient(
            host=self.host,
            port=self.port,
            **kwargs
        )

    def query(self, q):
        assert isinstance(q, dict)

        conn = self._get_connection()
        db = conn[self.database]
        coll = db[self.collection]

        raw_data = coll.find(q, {'_id': 0})
        if self.sort_by:
            raw_data = raw_data.sort(list((k, self.sort_by[k]) for k in self.sort_by))
        if self.limit:
            raw_data = raw_data.limit(self.limit)

        df = pd.DataFrame(list(raw_data))
        for col in df.columns:
            if not is_numeric_dtype(df[col]) or isinstance(df[col], dict):
                df[col] = df[col].astype(str)

        return df, self._make_colmap(df)

    def name(self):
        return 'MongoDB - {}'.format(self._query)

    def get_columns(self):
        ''' Return collection top level keys
            Note: 'reduce' is not work at Atlas
        '''
        conn = self._get_connection()
        db = conn[self.database]
        coll = db[self.collection]
        result = coll.aggregate(
            [ {
                "$match": self._query
            }, {
                "$project": {
                    "arrayofkeyvalue": {
                        "$objectToArray": "$$ROOT"
                    }
                }
            }, {
                "$unwind": "$arrayofkeyvalue"
            }, {
                "$group": {
                    "_id":None,
                    "allkeys": {
                        "$addToSet": "$arrayofkeyvalue.k"
                    }
                }
            } ]
        )
        result = list(result)[0]['allkeys']
        result.remove('_id')
        return result

    def get_row_count(self):
        conn = self._get_connection()
        db = conn[self.database]
        coll = db[self.collection]
        row_count = coll.find(self._query).count()
        return row_count



import os

import pandas as pd
from cassandra.cluster import Cluster
from cassandra.auth import PlainTextAuthProvider

from mindsdb_datasources.datasources.data_source import SQLDataSource


class ScyllaDS(SQLDataSource):
    ''' ScyllaDB use CQL, which pretty close to SQL, so filtering and other should work in main cases
        database == keyspace
    '''
    def __init__(self, query, database='', host='localhost', port=9042,
                 user='', password='', secure_connect_bundle=None, protocol_version=None):
        super().__init__(query)
        self.keyspace = database
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.secure_connect_bundle = secure_connect_bundle
        self.protocol_version = protocol_version

    def query(self, q):
        auth_provider = PlainTextAuthProvider(
            username=self.user, password=self.password
        )
        connection_props = {
            'auth_provider': auth_provider
        }

        if self.protocol_version is not None:
            connection_props['protocol_version'] = self.protocol_version

        if self.secure_connect_bundle is not None:
            if os.path.isfile(self.secure_connect_bundle) is False:
                raise Exception("'secure_connect_bundle' must be path to the file")
            connection_props['cloud'] = {
                'secure_connect_bundle': self.secure_connect_bundle
            }
        else:
            connection_props['contact_points'] = [self.host]
            connection_props['port'] = int(self.port)

        cluster = Cluster(**connection_props)
        session = cluster.connect()

        if isinstance(self.keyspace, str) and len(self.keyspace) > 0:
            session.set_keyspace(self.keyspace)

        resp = session.execute(q).all()

        df = pd.DataFrame(resp)

        df.columns = [x if isinstance(x, str) else x.decode('utf-8') for x in df.columns]
        for col_name in df.columns:
            try:
                df[col_name] = df[col_name].apply(lambda x: x if isinstance(x, str) else x.decode('utf-8'))
            except Exception:
                pass

        return df, self._make_colmap(df)

    def name(self):
        return 'ScyllaDB - {}'.format(self._query)



import os

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from mindsdb_datasources.datasources.data_source import DataSource, FileDS
from mindsdb_datasources import F


class S3DS(DataSource):
    def __init__(self, bucket_name, file_path, access_key=None,
                 secret_key=None, use_default_credentails=False):
        super().__init__()
        self.bucket_name = bucket_name
        self.file_path = file_path
        self.access_key = access_key
        self.secret_key = secret_key
        self.use_default_credentails = use_default_credentails

    def query(self, q):
        if self.access_key is not None and secret_key is not None:
            s3 = boto3.client('s3', aws_access_key_id=self.access_key, aws_secret_access_key=self.secret_key)
        elif use_default_credentails:
            s3 = boto3.client('s3')
        else:
            s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))

        with open(self.tmp_file_name, 'wb') as fw:
            s3.download_fileobj(self.bucket_name, self.file_path, fw)

        tmp_file_name = '.tmp_mindsdb_data_file'

        file_ds = FileDS(tmp_file_name)

        os.remove(tmp_file_name)

        return file_ds.df, file_ds._col_map

    def name(self):
        return 'S3 - {}/{}'.format(self.bucket_name, os.path.basename(self.file_path))



import pandas as pd
import phoenixdb

from mindsdb_datasources.datasources.data_source import SQLDataSource


class PhoenixDS(SQLDataSource):
    def __init__(self, query, url, auth=None, authentication=None, user=None,
                 password=None, verify=None):
        super().__init__(query=query)
        self.url = url
        self.auth = auth
        self.authentication = authentication
        self.user = user
        self.password = password
        self.verify = verify

    def query(self, q):
        con = phoenixdb.connect(
            url=self.url,
            auth=self.auth,
            authentication=self.authentication,
            user=self.user,
            password=self.password,
            verify=self.verify
        )

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Phoenix - {}'.format(self._query)



import unittest
from common import DB_CREDENTIALS, break_dataset


class TestScylla(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['scylla']['user']
        self.PASSWORD = DB_CREDENTIALS['scylla']['password']
        self.HOST = DB_CREDENTIALS['scylla']['host']
        self.PORT = int(DB_CREDENTIALS['scylla']['port'])
        self.KEYSPACE = 'test_data'
        self.TABLE = 'home_rentals'

    def test_scylla_ds(self):
        from mindsdb_datasources import ScyllaDS

        LIMIT = 100

        scylla_ds = ScyllaDS(
            host=self.HOST,
            user=self.USER,
            password=self.PASSWORD,
            database=self.KEYSPACE,
            port=self.PORT,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        scylla_ds.df = break_dataset(scylla_ds.df)

        assert len(scylla_ds) == LIMIT




import pandas as pd
from sqlite3 import connect

from mindsdb_datasources.datasources.data_source import SQLDataSource

class SQLite3DS(SQLDataSource):
    def __init__(self, query, database):
        super().__init__(query=query)
        self.database = database

    def query(self, q):
        with connect(self.database) as con:
            df = pd.read_sql(q, con=con)
        return df, self._make_colmap(df)
    
    def name(self):
        return 'SQLite3 - {}'.format(self._query)


import os

from google.cloud import storage
from google.oauth2 import service_account
import json

from mindsdb_datasources.datasources.data_source import DataSource, FileDS
from mindsdb_datasources import F


class GCSDS(DataSource):
    def __init__(self, bucket_name, file_path, project_id, auth_json):
        super().__init__()
        self.bucket_name = bucket_name
        self.file_path = file_path
        self.auth_json = auth_json
        self.project_id = project_id

    def query(self, q=None):
        try:
            auth_info = json.loads(self.auth_json)
            credentials = service_account.Credentials.from_service_account_info(info=auth_info)
        except Exception:
            credentials = service_account.Credentials.from_service_account_file(filename=self.auth_json)

        gc_client = storage.Client(credentials=credentials, project=self.project_id)
        bucket = gc_client.get_bucket(self.bucket_name)
        blob = storage.Blob(self.file_path, bucket)

        self.tmp_file_name = '.tmp_mindsdb_data_file'

        with open(self.tmp_file_name, 'wb') as fw:
            gc_client.download_blob_to_file(blob, fw)

        file_ds = FileDS(self.tmp_file_name)

        os.remove(self.tmp_file_name)

        return file_ds.df, file_ds._col_map

    def name(self):
        return 'GCS - {}/{}'.format(self.bucket_name, os.path.basename(self.file_path))



import pandas as pd
from pyhive import hive

from mindsdb_datasources.datasources.data_source import SQLDataSource


class HiveDS(SQLDataSource):
    def __init__(self, query, host='localhost', port=10000,
                 database='default',  username=None, password=None,
                 auth=None, configuration=None, kerberos_service_name=None,
                 check_hostname=None, scheme='http', ssl_cert=None):

        super().__init__(query=query)
        self.host = host
        self.port = int(port)
        self.database = database
        self.username = username
        self.password = password
        self.auth = auth
        self.configuration = configuration
        self.kerberos_service_name = kerberos_service_name
        self.check_hostname = check_hostname
        self.scheme = scheme
        self.ssl_cert = ssl_cert

    def query(self, q):
        con = hive.connect(
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
            auth=self.auth,
            configuration=self.configuration,
            kerberos_service_name=self.kerberos_service_name,
            check_hostname=self.check_hostname,
            scheme=self.scheme,
            ssl_cert=self.ssl_cert,
        )

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Hive - {}'.format(self._query)



import unittest
from common import DB_CREDENTIALS, break_dataset


class TestMYSQL(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['mysql']['user']
        self.PASSWORD = DB_CREDENTIALS['mysql']['password']
        self.HOST = DB_CREDENTIALS['mysql']['host']
        self.PORT = int(DB_CREDENTIALS['mysql']['port'])
        self.DATABASE = 'test_data'
        self.TABLE = 'us_health_insurance'

    def test_mysql_ds(self):
        from mindsdb_datasources import MySqlDS

        LIMIT = 400

        mysql_ds = MySqlDS(
            host=self.HOST,
            user=self.USER,
            password=self.PASSWORD,
            database=self.DATABASE,
            port=self.PORT,
            query=' (SELECT * FROM (SELECT * FROM {table} LIMIT {limit}) as t1) UNION ALL (SELECT * FROM (SELECT * FROM {table} LIMIT {limit}) as t1)'.format(table=self.TABLE, limit=int(LIMIT/2))
        )

        mysql_ds.df = break_dataset(mysql_ds.df)

        assert len(mysql_ds) <= LIMIT

        # Our SQL parsing fails here, test if we're still able to filter via the dataframe fallback
        for val in mysql_ds.filter([['sex', 'like','fem']])['sex']:
            assert val == 'female'

        assert len(mysql_ds.filter([['age', '>', 20]], 12)) == 12
        assert len(mysql_ds.filter([['age', '=', 60]], 1)) == 1
        assert len(mysql_ds.filter([['age', '>', 150]], 11)) == 0



import unittest
from common import DB_CREDENTIALS, break_dataset


class TestMariaDB(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['mariadb']['user']
        self.PASSWORD = DB_CREDENTIALS['mariadb']['password']
        self.HOST = DB_CREDENTIALS['mariadb']['host']
        self.PORT = int(DB_CREDENTIALS['mariadb']['port'])
        self.DATABASE = 'test_data'
        self.TABLE = 'hdi'

    def test_maria_ds(self):
        from mindsdb_datasources import MariaDS

        LIMIT = 200

        maria_ds = MariaDS(
            host=self.HOST,
            user=self.USER,
            password=self.PASSWORD,
            database=self.DATABASE,
            port=self.PORT,
            query='SELECT * FROM `{}` LIMIT {}'.format(self.TABLE, LIMIT)
        )

        maria_ds.df = break_dataset(maria_ds.df)

        assert len(maria_ds) <= LIMIT

        # Our SQL parsing succeds here, but the query fails, test if we're still able to filter via the dataframe fallback
        maria_ds._query = maria_ds._query.replace(self.TABLE, 'wrongly_named_table')
        assert len(maria_ds.filter([['Population', '<', 33098932]], 8)) == 8
        assert len(maria_ds.filter([['Development_Index', '!=', 3]], 12)) == 12



import os
import unittest
import logging
from common import DB_CREDENTIALS, break_dataset


class TestMSSQL(unittest.TestCase):
    def test_mssql_ds(self):
        from mindsdb_datasources import MSSQLDS

        HOST = DB_CREDENTIALS['mssql']['host']
        USER = DB_CREDENTIALS['mssql']['user']
        PASSWORD = DB_CREDENTIALS['mssql']['password']
        DATABASE = DB_CREDENTIALS['mssql']['database']
        PORT = DB_CREDENTIALS['mssql']['port']

        mssql_ds = MSSQLDS(
            query='SELECT * FROM dbo.insurance LIMIT',
            host=HOST,
            user=USER,
            password=PASSWORD,
            database=DATABASE,
            port=PORT
        )

        assert (len(mssql_ds.df) > 200)



import os

import pandas as pd
import pg8000

from mindsdb_datasources.datasources.data_source import SQLDataSource


class RedshiftDS(SQLDataSource):
    def __init__(self, query, database='dev', host='localhost',
                 port=5439, user='awsuser', password=''):
        super().__init__(query=query)
        self.database = database
        self.host = host
        self.port = int(port)
        self.user = user
        self.password = password

    def query(self, q):
        con = pg8000.connect(
            database=self.database,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port
        )

        df = pd.read_sql(q, con=con)
        con.close()

        for col_name in df.columns:
            try:
                df[col_name] = df[col_name].apply(lambda x: x.decode("utf-8"))
            except Exception:
                pass

        return df, self._make_colmap(df)

    def name(self):
        return 'Redshift - {}'.format(self._query)



from mindsdb_datasources.datasources.data_source import DataSource, SQLDataSource
from mindsdb_datasources.datasources.maria_ds import MariaDS
from mindsdb_datasources.datasources.mysql_ds import MySqlDS
from mindsdb_datasources.datasources.clickhouse_ds import ClickhouseDS
from mindsdb_datasources.datasources.file_ds import FileDS
from mindsdb_datasources.datasources.sqlite3_ds import SQLite3DS


from mindsdb_datasources.__about__ import __package_name__ as name, __version__

failed_import_ds = []

# These might not initialized properly since they require optional dependencies, so we wrap them in a try-except
try:
    from mindsdb_datasources.datasources.snowflake_ds import SnowflakeDS
except ImportError:
    failed_import_ds.append('Snowflake')
    SnowflakeDS = None

try:
    from mindsdb_datasources.datasources.s3_ds import S3DS
except:
    failed_import_ds.append('S3')
    S3DS = None

try:
    from mindsdb_datasources.datasources.postgres_ds import PostgresDS
except ImportError:
    failed_import_ds.append('PostgreSQL')
    PostgresDS = None

try:
    from mindsdb_datasources.datasources.ms_sql_ds import MSSQLDS
except ImportError:
    failed_import_ds.append('MSSQL')
    MSSQLDS = None

try:
    from mindsdb_datasources.datasources.mongodb_ds import MongoDS
except ImportError:
    failed_import_ds.append('MongoDB')
    MongoDS = None

try:
    from mindsdb_datasources.datasources.aws_athena_ds import AthenaDS
except ImportError:
    failed_import_ds.append('Athena')
    AthenaDS = None

try:
    from mindsdb_datasources.datasources.redshift_ds import RedshiftDS
except ImportError:
    failed_import_ds.append('Redshift')
    RedshiftDS = None

try:
    from mindsdb_datasources.datasources.gcs_ds import GCSDS
except ImportError:
    failed_import_ds.append('Google Cloud Storage')
    GCSDS = None

try:
    from mindsdb_datasources.datasources.scylla_ds import ScyllaDS
except ImportError:
    failed_import_ds.append('ScyllaDB')
    ScyllaDS = None

try:
    from mindsdb_datasources.datasources.cassandra_ds import CassandraDS
except ImportError:
    failed_import_ds.append('Cassandra')
    CassandraDS = None

try:
    from mindsdb_datasources.datasources.solr_ds import SolrDS
except ImportError:
    failed_import_ds.append('Solr')
    SolrDS = None
    
try:    
    from mindsdb_datasources.datasources.presto_ds import PrestoDS
except:
    failed_import_ds.append('Presto')
    PrestoDS = None

try:
    from mindsdb_datasources.datasources.timescale_ds import TimescaleDS
except ImportError:
    failed_import_ds.append('Timescale')
    TimescaleDS = None

try:
    from mindsdb_datasources.datasources.dremio_ds import DremioDS
except ImportError:
    failed_import_ds.append('Dremio')
    DremioDS = None

try:
    from mindsdb_datasources.datasources.bigquery_ds import BigQueryDS
except ImportError:
    failed_import_ds.append('BigQuery')
    BigQueryDS = None

try:
    from mindsdb_datasources.datasources.impala_ds import ImpalaDS
except ImportError:
    failed_import_ds.append('Impala')
    ImpalaDS = None

try:
    from mindsdb_datasources.datasources.hive_ds import HiveDS
except ImportError:
    failed_import_ds.append('Hive')
    HiveDS = None

try:
    from mindsdb_datasources.datasources.trino_ds import TrinoDS
except ImportError:
    failed_import_ds.append('Trino')
    TrinoDS = None

try:
    from mindsdb_datasources.datasources.influx_ds import InfluxDS
except ImportError:
    failed_import_ds.append('InfluxDB')
    InfluxDS = None

try:
    from mindsdb_datasources.datasources.phoenix_ds import PhoenixDS
except ImportError:
    failed_import_ds.append('Phoenix')
    PhoenixDS = None

try:
    from mindsdb_datasources.datasources.quest_ds import QuestDS
except ImportError:
    failed_import_ds.append('QuestDB')
    QuestDS = None

try:
    from mindsdb_datasources.datasources.crate_ds import CrateDS
except ImportError:
    failed_import_ds.append('CrateDB')
    CrateDS = None

if len(failed_import_ds) > 0:
    print('Some datasources are not available by default. How to install them, read here: https://github.com/mindsdb/datasources#installing-additional-dependencies')



import unittest
from common import DB_CREDENTIALS
from mindsdb_datasources import MongoDS


class TestMongoDB(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['mongodb']['user']
        self.PASSWORD = DB_CREDENTIALS['mongodb']['password']
        self.HOST = DB_CREDENTIALS['mongodb']['host']
        self.PORT = int(DB_CREDENTIALS['mongodb']['port'])
        self.DATABASE = 'test_data'
        self.COLLECTION = 'home_rentals'

    def test_mongodb_ds(self):

        mongodb_ds = MongoDS(
            collection=self.COLLECTION,
            query={},
            host=self.HOST,
            port=self.PORT,
            user=self.USER,
            password=self.PASSWORD,
            database=self.DATABASE
        )

        for val in mongodb_ds.filter([['location', 'like','ood']])['location']:
            assert val == 'good'

        assert len(mongodb_ds.filter([['rental_price', '>', 2500]], 3)) == 3
        assert len(mongodb_ds.filter([['initial_price', '<', 0]], 3)) == 0

    # def test_mongodb_atlas_ds(self):
    #     USER = DB_CREDENTIALS['mongodb_atlas']['user']
    #     PASSWORD = DB_CREDENTIALS['mongodb_atlas']['password']
    #     HOST = DB_CREDENTIALS['mongodb_atlas']['host']
    #     PORT = int(DB_CREDENTIALS['mongodb_atlas']['port'])
    #     DATABASE = 'sample_restaurants'
    #     COLLECTION = 'restaurants'
    #     atlas_ds = MongoDS(
    #         collection=COLLECTION,
    #         query={"grades": []},
    #         host=HOST,
    #         port=PORT,
    #         user=USER,
    #         password=PASSWORD,
    #         database=DATABASE,
    #         limit=5,
    #     )
    #     assert len(atlas_ds.df) == 5

    #     df, _ = atlas_ds.query({"grades":{"$size": 5}})
    #     assert len(df) == 5
    #     df, _ = atlas_ds.query({"grades":{"$size": 45}})
    #     assert len(df) == 0



import pandas as pd
from influxdb import InfluxDBClient

from mindsdb_datasources.datasources.data_source import SQLDataSource


class InfluxDS(SQLDataSource):
    def __init__(self,query,host, port, 
                username, password, ssl, verify_ssl):
        super().__init__(query)
        self.host = host
        self.port = int(port)
        self.username = username
        self.password = password
        self.ssl = ssl
        self.verify_ssl = verify_ssl
    
    def query(self,q):
        client = InfluxDBClient(host=self.host, port=self.port, username=self.username, 
                                password=self.password, ssl=self.ssl, verify_ssl=self.verify_ssl)

        result = client.query(q)
        result = result.raw
        df = pd.json_normalize(result)

        df.columns = [x if isinstance(x, str) else x.decode('utf-8') for x in df.columns]
        for col_name in df.columns:
            try:
                df[col_name] = df[col_name].apply(lambda x: x if isinstance(x, str) else x.decode('utf-8'))
            except Exception:
                pass

        return df, self._make_colmap(df)




import time
import random
import string
import unittest
import requests
from common import DB_CREDENTIALS, break_dataset
import gc


def random_string():
    return ''.join(
        random.Random(
            int(time.time() * 1e6)
        ).choices(
            string.ascii_letters + string.digits,
            k=6
        )
    )


class TestClickhouse(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['clickhouse']['user']
        self.PASSWORD = DB_CREDENTIALS['clickhouse']['password']
        self.HOST = DB_CREDENTIALS['clickhouse']['host']
        self.PORT = int(DB_CREDENTIALS['clickhouse']['port'])
        self.DATABASE = 'test_data'


    def test_clickhouse_ds(self):
        from mindsdb_datasources import ClickhouseDS
        LIMIT = 100

        clickhouse_ds = ClickhouseDS(
            host=self.HOST,
            port=self.PORT,
            user=self.USER,
            password=self.PASSWORD,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                self.DATABASE,
                'home_rentals',
                LIMIT
            )
        )

        # test filter
        for val in clickhouse_ds.filter([['location', 'like','ood']])['location']:
            assert val == 'good'

        assert len(clickhouse_ds.filter([['rental_price', '>', 2500]], 3)) == 3
        assert len(clickhouse_ds.filter([['initial_price', '<', 0]], 3)) == 0

        # mess with the values inside then try to analyze it
        clickhouse_ds.df = break_dataset(clickhouse_ds.df)
        assert len(clickhouse_ds) <= LIMIT

    @unittest.skip("Skip untill we move to Lightwood")
    def test_database_history(self):
        return
        from mindsdb_datasources import ClickhouseDS

        TEMP_DB = 'test_database_history_' + random_string()
        TEMP_TABLE = 'tmp_test_database_history_' + random_string()

        params = {'user': self.USER, 'password': self.PASSWORD}

        clickhouse_url = f'http://{self.HOST}:{self.PORT}'

        values = []
        for i in range(200):
            values.append([str(i % 4), i, i * 2])

        queries = [
            f'CREATE DATABASE IF NOT EXISTS {TEMP_DB}',
            f'DROP TABLE IF EXISTS {TEMP_DB}.{TEMP_TABLE}',
            f'''
                CREATE TABLE {TEMP_DB}.{TEMP_TABLE}(
                    col1 String
                    ,col2 Int64
                    ,col3 Int64
                ) ENGINE = MergeTree()
                    ORDER BY col2
                    PARTITION BY col1
            ''',
        ]
        gc.collect()

        for value in values:
            value_ins_str = str(value).replace('[','').replace(']','')
            queries.append(f"INSERT INTO {TEMP_DB}.{TEMP_TABLE} VALUES ({value_ins_str})")

        for q in queries:
            r = requests.post(clickhouse_url, data=q, params=params)
            assert r.status_code == 200, r.text

        clickhouse_ds = ClickhouseDS(
            f'SELECT * FROM {TEMP_DB}.{TEMP_TABLE}',
            host=self.HOST,
            port=self.PORT,
            user=self.USER,
            password=self.PASSWORD
        )

        temp_predictor = Predictor(name='query_history_based_ts_predictor')
        temp_predictor.learn(
            to_predict='col3',
            from_data=clickhouse_ds,
            stop_training_in_x_seconds=5,
            timeseries_settings={
                'order_by': ['col2']
                ,'window': 6
                ,'group_by': ['col1']
            }
        )
        del temp_predictor

        ts_predictor = Predictor(name='query_history_based_ts_predictor')
        predictions = ts_predictor.predict(
            when_data={
                'col2': 800
                ,'col1': '2'
            },
            advanced_args={
                'use_database_history': True
            }
        )

        assert predictions[0]['col3'] is not None

        r = requests.post(
            clickhouse_url,
            data=f'DROP DATABASE {TEMP_DB}',
            params=params
        )
        assert r.status_code == 200, 'failed to drop temporary database "{}"'.format(TEMP_DB)



import unittest
from common import DB_CREDENTIALS, break_dataset

@unittest.skip("Ignore untill we have environment ready")
class TestHive(unittest.TestCase):
    def setUp(self):
        self.HOST = DB_CREDENTIALS['hive']['host']
        self.PORT = DB_CREDENTIALS['hive']['port']
        self.DATABASE = 'hive'

    def test_hive_ds(self):
        from mindsdb_datasources import HiveDS

        LIMIT = 100

        hive_ds = HiveDS(
            host=self.HOST,
            port=self.PORT,
            database=self.DATABASE,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        hive_ds.df = break_dataset(hive_ds.df)

        assert len(hive_ds) == LIMIT



import pandas as pd
import mysql.connector

from mindsdb_datasources.datasources.data_source import SQLDataSource


class MariaDS(SQLDataSource):
    def __init__(self, query, database='mysql', host='localhost',
                 port=3306, user='root', password='',
                 ssl=None, ssl_ca=None, ssl_cert=None, ssl_key=None):
        super().__init__(query)
        self.database = database
        self.host = host
        self.port = int(port)
        self.user = user
        self.password = password
        self.ssl = ssl
        self.ssl_ca = ssl_ca
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key

    def query(self, q):
        config = {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
            "database": self.database
        }
        if self.ssl is True:
            config['client_flags'] = [mysql.connector.constants.ClientFlag.SSL]
            if self.ssl_ca is not None:
                config["ssl_ca"] = self.ssl_ca
            if self.ssl_cert is not None:
                config["ssl_cert"] = self.ssl_cert
            if self.ssl_key is not None:
                config["ssl_key"] = self.ssl_key
        con = mysql.connector.connect(**config)

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'MariaDB - {}'.format(self._query)



from mindsdb_datasources.datasources.data_source import SQLDataSource
import pandas as pd
import pyodbc


class SnowflakeDS(SQLDataSource):
    def __init__(self, query, host, user, password, account, warehouse,
                 database, schema, protocol='https', port=443):
        super().__init__(query)
        self.host = host
        self.user = user
        self.password = password
        self.account = account
        self.warehouse = warehouse
        self.database = database
        self.schema = schema
        self.protocol = protocol
        self.port = int(port)

    def query(self, q):
        print(self.host)
        con = pyodbc.connect(f"Driver={{SnowflakeDSIIDriver}}; Server={self.host}; Database={self.database}; schema={self.schema}; UID={self.user}; PWD={self.password}; warehouse={self.warehouse}; protocol={self.protocol}; port={self.port}; account={self.account}")

        cur = con.cursor().execute(q)
        columns = [column[0] for column in cur.description]
        results = []
        for row in cur.fetchall():
            results.append(dict(zip(columns, row)))
        df = pd.DataFrame(results)
        cur.close()
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Snowflake - {}'.format(self._query)



import impala.dbapi
from impala.util import as_pandas
import pandas as pd

from mindsdb_datasources.datasources.data_source import SQLDataSource


class ImpalaDS(SQLDataSource):
    def __init__(self, query, host='localhost', port=21050, use_ssl=None,
                 ca_cert=None, auth_mechanism='NOSASL', user=None,
                 password=None, database=None):
        super().__init__(query=query)
        self.host = host
        self.port = int(port)
        self.use_ssl = use_ssl
        self.ca_cert = ca_cert
        self.auth_mechanism = auth_mechanism
        self.user = user
        self.password = password
        self.database = database

    def query(self, q):
        con = impala.dbapi.connect(
            host=self.host,
            port=self.port,
            use_ssl=self.use_ssl,
            ca_cert=self.ca_cert,
            auth_mechanism=self.auth_mechanism,
            user=self.user,
            password=self.password,
            database=self.database
        )
        cur = con.cursor()
        cur.execute(q)

        df = as_pandas(cur)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Impala - {}'.format(self._query)



from pyathena import connect
from pyathena.pandas.util import as_pandas

from mindsdb_datasources.datasources.data_source import SQLDataSource


class AthenaDS(SQLDataSource):
    def __init__(self, query, staging_dir, database=None,
                 access_key=None, secret_key=None, region_name=None, table=None):
        """
        :param query: Query to be executed. Ex. SELECT * FROM db.table;
        :param staging_dir: Full S3 path where Athena temp data will stored. Ex. s3://bucket_name/athena/staging
        :param database: Name of the Database
        :param table: Name of the Table
        :param access_key: Access Key used if supplied else used default credentials.
        :param secret_key: Secret Key used if supplied else used default credentials.
        :param region_name: Region used if supplied else used default region.
        """
        super().__init__(query)

        if (not database or not table) and not query:
            raise ValueError('Either database and table or query should be passed.')
        
        self.staging_dir = staging_dir
        self.database = database
        self.access_key = access_key
        self.secret_key = secret_key
        self.region_name = region_name

    def query(self, q):
        _conn_args = {
            's3_staging_dir': self.staging_dir,
            'database': self.database
        }

        if self.access_key is not None and self.secret_key is not None:
            _conn_args['aws_access_key_id'] = self.access_key
            _conn_args['aws_secret_access_key'] = self.secret_key

        if self.region_name is not None:
            _conn_args['region_name'] = self.region_name

        conn = connect(**_conn_args)
        cursor = conn.cursor()

        cursor.execute(q)

        # Load query results into Pandas DataFrame and show results
        df = as_pandas(cursor)

        col_map = {}
        for col in df.columns:
            col_map[col] = col

        return df, col_map

    def name(self):
        return 'Athena - {}'.format(self._query)



import unittest
from common import DB_CREDENTIALS, break_dataset

@unittest.skip("Ignore untill we have environment ready")
class TestImpala(unittest.TestCase):
    def setUp(self):
        self.HOST = DB_CREDENTIALS['impala']['host']
        self.PORT = int(DB_CREDENTIALS['impala']['port'])
        self.DATABASE = 'impala'

    def test_impala_ds(self):
        from mindsdb_datasources import ImpalaDS

        LIMIT = 100

        impala_ds = ImpalaDS(
            host=self.HOST,
            port=self.PORT,
            database=self.DATABASE,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        impala_ds.df = break_dataset(impala_ds.df)

        assert len(impala_ds) == LIMIT



import os
import json
import time
import random
import string
import unittest
import numpy as np

_var_name = 'DATABASE_CREDENTIALS_STRINGIFIED_JSON'
_var_value = os.getenv(_var_name)
if _var_value is None:
    with open(os.path.join(os.path.expanduser("~"), '.mindsdb_credentials.json'), 'r') as fp:
        _var_value = fp.read()

assert _var_value is not None, _var_name + ' ' + 'is not set'


DB_CREDENTIALS = json.loads(_var_value)


def break_dataset(df, null_cell_pct=0.2, stringify_numbers_pct=0.5):
    df = df.copy()

    n_rows, n_cols = df.shape

    # Make cells null
    for i in range(n_rows):
        for j in range(n_cols):
            if random.random() <= null_cell_pct:
                df.iloc[i, j] = np.nan

    # Stringify numbers
    for i in range(n_rows):
        for j in range(n_cols):
            if isinstance(df.iloc[i, j], (float, int)):
                if random.random() <= stringify_numbers_pct:
                    df.iloc[i, j] = str(df.iloc[i, j])

    return df



import os
import unittest
#from common import DB_CREDENTIALS

@unittest.skip("Ignore untill we have environment ready")
class TestDremioDS(unittest.TestCase):
    def test_dremio_ds(self):
        from mindsdb_datasources import DremioDS

        LIMIT = 100

        dremio_ds = DremioDS(
            query=f'SELECT * FROM foo.bar LIMIT {LIMIT}',
            host=DB_CREDENTIALS['dremio']['host'],
            user=DB_CREDENTIALS['dremio']['port'],
            password=DB_CREDENTIALS['dremio']['user'],
        )

        assert len(dremio_ds.df) == LIMIT




from mindsdb_datasources.datasources.scylla_ds import ScyllaDS


class CassandraDS(ScyllaDS):
    ''' Cassandra and ScyllaDB very similar.
        Differ only drivers, but connection can be established with both.
    '''
    def name(self):
        return 'CassandraDB - {}'.format(self._query)



import os
from io import BytesIO, StringIO
import csv
import codecs
import json
import traceback
import shutil
from urllib.parse import urlparse
import tempfile

import pandas as pd
import requests

from mindsdb_datasources.datasources.data_source import DataSource


def clean_row(row):
    n_row = []
    for cell in row:
        if str(cell) in ['', ' ', '  ', 'NaN', 'nan', 'NA']:
            n_row.append(None)
        else:
            n_row.append(cell)

    return n_row


accepted_csv_delimiters = [',', '\t', ';']


class FileDS(DataSource):
    def __init__(self, file, clean_rows=True, custom_parser=None):
        """
        Setup from file
        :param file: fielpath or url
        :param clean_rows: if you want to clean rows for strange null values
        :param custom_parser: if you want to parse the file with some custom parser
        """
        if not isinstance(file, str):
            raise ValueError("'file' must be string")
        super().__init__()
        self.file = file
        self.clean_rows = clean_rows
        self.custom_parser = custom_parser
        self.dialect = None

        self._column_names = None
        self._row_count = None

        self._temp_dir = None

        try:
            self.is_url = urlparse(file).scheme in ('http', 'https')
        except Exception:
            self.is_url = False

    def __del__(self):
        if self._temp_dir is not None:
            shutil.rmtree(self._temp_dir)

    def _handle_source(self):
        self._file_name = os.path.basename(self.file)

        # get file data io, format and dialect
        data, fmt, self.dialect = self._getDataIo(self.file)
        data.seek(0)  # make sure we are at 0 in file pointer

        if self.custom_parser:
            header, file_data = self.custom_parser(data, fmt)

        elif fmt == 'csv':
            csv_reader = list(csv.reader(data, self.dialect))
            header = csv_reader[0]
            file_data = csv_reader[1:]

        elif fmt in ['xlsx', 'xls']:
            data.seek(0)
            df = pd.read_excel(data)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        elif fmt == 'json':
            data.seek(0)
            json_doc = json.loads(data.read())
            df = pd.json_normalize(json_doc, max_level=0)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        else:
            raise ValueError('Could not load file into any format, supported formats are csv, json, xls, xlsx')

        if self.clean_rows:
            file_list_data = [clean_row(row) for row in file_data]
        else:
            file_list_data = file_data

        col_map = dict((col, col) for col in header)
        return pd.DataFrame(file_list_data, columns=header), col_map

    def query(self, q=None):
        try:
            return self._handle_source()
        except Exception as e:
            print(f"Error creating dataframe from handled data: {e}")
            print("pd.read_csv data handler would be used.")
            df = pd.read_csv(self.file, sep=self.dialect.delimiter)
            col_map = dict((col, col) for col in df.columns)
            return df, col_map

    def _getDataIo(self, file):
        """
        This gets a file either url or local file and defiens what the format is as well as dialect
        :param file: file path or url
        :return: data_io, format, dialect
        """

        ############
        # get file as io object
        ############

        file_path = self._get_file_path()

        data = BytesIO()

        try:
            with open(file_path, 'rb') as fp:
                data = BytesIO(fp.read())
        except Exception as e:
            error = 'Could not load file, possible exception : {exception}'.format(exception=e)
            print(error)
            raise ValueError(error)

        dialect = None

        ############
        # check for file type
        ############

        # try to guess if its an excel file
        xlsx_sig = b'\x50\x4B\x05\06'
        # xlsx_sig2 = b'\x50\x4B\x03\x04'
        xls_sig = b'\x09\x08\x10\x00\x00\x06\x05\x00'

        # different whence, offset, size for different types
        excel_meta = [('xls', 0, 512, 8), ('xlsx', 2, -22, 4)]

        for filename, whence, offset, size in excel_meta:

            try:
                data.seek(offset, whence)  # Seek to the offset.
                bytes = data.read(size)  # Capture the specified number of bytes.
                data.seek(0)
                codecs.getencoder('hex')(bytes)

                if bytes == xls_sig:
                    return data, 'xls', dialect
                elif bytes == xlsx_sig:
                    return data, 'xlsx', dialect

            except Exception:
                data.seek(0)

        # if not excel it can be a json file or a CSV, convert from binary to stringio

        byte_str = data.read()
        # Move it to StringIO
        try:
            # Handle Microsoft's BOM "special" UTF-8 encoding
            if byte_str.startswith(codecs.BOM_UTF8):
                data = StringIO(byte_str.decode('utf-8-sig'))
            else:
                data = StringIO(byte_str.decode('utf-8'))

        except Exception:
            print(traceback.format_exc())
            print('Could not load into string')

        # see if its JSON
        buffer = data.read(100)
        data.seek(0)
        text = buffer.strip()
        # analyze first n characters
        if len(text) > 0:
            text = text.strip()
            # it it looks like a json, then try to parse it
            if text.startswith('{') or text.startswith('['):
                try:
                    json.loads(data.read())
                    data.seek(0)
                    return data, 'json', dialect
                except Exception:
                    data.seek(0)
                    return data, None, dialect

        # lets try to figure out if its a csv
        try:
            dialect = self._get_csv_dialect()
            if dialect:
                return data, 'csv', dialect
            return data, None, dialect
        except Exception:
            data.seek(0)
            print('Could not detect format for this file')
            print(traceback.format_exc())
            # No file type identified
            return data, None, dialect

    def name(self):
        return 'File, {}'.format(os.path.basename(self.file))

    def _fetch_url(self):
        if self._temp_dir is not None and os.path.isfile(os.path.join(self._temp_dir, 'file')):
            return
        self._temp_dir = tempfile.mkdtemp(prefix='mindsdb_fileds_')
        try:
            r = requests.get(self.file, stream=True)
            if r.status_code == 200:
                with open(os.path.join(self._temp_dir, 'file'), 'wb') as f:
                    for chunk in r:
                        f.write(chunk)
            else:
                raise Exception(f'Responce status code is {r.status_code}')
        except Exception as e:
            print(f'Error during getting {self.file}')
            print(e)
            raise

    def _get_csv_dialect(self) -> object:
        file_path = self._get_file_path()
        with open(file_path, 'rt') as f:
            try:
                dialect = csv.Sniffer().sniff(f.read(128 * 1024), delimiters=accepted_csv_delimiters)
            except csv.Error:
                dialect = None
        return dialect

    def _read_meta(self):
        ''' determine 'column_names' and 'row_count' for DS

            TODO: at this moment optimisation exists only for .csv files,
            all other files will be read whole, which is slow
        '''
        dialect = self._get_csv_dialect()
        if dialect is not None:
            file_path = self._get_file_path()
            with open(file_path) as csvfile:
                csv_reader = csv.reader(csvfile, delimiter=dialect.delimiter)
                self._column_names = next(csv_reader, None)
                self._row_count = sum(1 for line in csv_reader)
        else:
            df = self.df
            self._column_names = list(df.keys())
            self._row_count = len(df)

    def _get_file_path(self) -> str:
        path = self.file
        if self.is_url:
            self._fetch_url()
            path = os.path.join(self._temp_dir, 'file')
        return path

    def get_columns(self) -> list:
        if self._column_names is None:
            self._read_meta()
        return self._column_names

    def get_row_count(self) -> int:
        if self._row_count is None:
            self._read_meta()
        return int(self._row_count)



from google.cloud import bigquery
from mindsdb_datasources.datasources.data_source import SQLDataSource


class BigQueryDS(SQLDataSource):
    def __init__(self, query, project=None, credentials=None,
                 location=None, default_query_job_config=None,
                 client_info=None, client_options=None):
        super().__init__(query)
        self.project = project
        self.credentials = credentials
        self.location = location
        self.default_query_job_config = default_query_job_config
        self.client_info = client_info
        self.client_options = client_options

    def query(self, q):
        client = bigquery.Client(
            project=self.project,
            credentials=self.credentials,
            location=self.location,
            default_query_job_config=self.default_query_job_config,
            client_info=self.client_info,
            client_options=self.client_options
        )
        
        df = client.query(q).to_dataframe()
        client.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Google BigQuery - {}'.format(self._query)




import pandas as pd
import prestodb

from mindsdb_datasources.datasources.data_source import SQLDataSource


class PrestoDS(SQLDataSource):
    def __init__(self, query, database, schema, host='localhost',
                 port=8080, user=None, password=None, protocol='https', catalog=None):
        super().__init__(query)
        self.database = database
        self.schema = schema
        self.host = host
        self.port = int(port)
        self.user = user
        self.password = password
        self.protocol = protocol
        self.catalog = catalog

    def query(self, q):
        con = prestodb.dbapi.connect(
            catalog=self.catalog,
            schema=self.schema,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            http_scheme=self.protocol
        )

        cur = con.cursor().execute(q)
        columns = [column[0] for column in cur.description]
        results = []
        for row in cur.fetchall():
            results.append(dict(zip(columns, row)))
        df = pd.DataFrame(results)
        cur.close()
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Presto - {}'.format(self._query)



import unittest
from common import DB_CREDENTIALS, break_dataset

@unittest.skip("Ignore untill we have environment ready")
class TestQuestDB(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['questdb']['user']
        self.PASSWORD = DB_CREDENTIALS['questdb']['password']
        self.HOST = DB_CREDENTIALS['questdb']['host']
        self.PORT = int(DB_CREDENTIALS['questdb']['port'])
        self.DATABASE = 'questdb'
        self.TABLE = 'test_table'

    def test_postgres_ds(self):
        from mindsdb_datasources import QuestDS

        LIMIT = 100

        quest_ds = QuestDS(
            host=self.HOST,
            user=self.USER,
            password=self.PASSWORD,
            database=self.DATABASE,
            port=self.PORT,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        quest_ds.df = break_dataset(quest_ds.df)

        assert len(quest_ds) == LIMIT



import pandas as pd
import pyodbc
from mindsdb_datasources.datasources.data_source import SQLDataSource


class DremioDS(SQLDataSource):
    def __init__(self, query, host, user, password, 
                 driver='Dremio ODBC Driver 64-bit', port=31010):
        super().__init__(query)
        self.host = host
        self.port = int(port)
        self.driver = driver
        self.user = user
        self.password = password

    def query(self, q):
        con = pyodbc.connect(
            f'Driver={self.driver};'
            f'HOST={self.host};'
            f'PORT={self.port};'
            f'UID={self.user};'
            f'PWD={self.password};'
            'AuthenticationType=Plain;'
            'ConnectionType=Direct;'
        )

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Dremio - {}'.format(self._query)




import pandas as pd
import jaydebeapi
from mindsdb_datasources.datasources.data_source import SQLDataSource

class SolrDS(SQLDataSource):
    def __init__(self, query, collection, user, password, host='localhost', port=9983):
        super().__init__(query=query)
        self.url = 'jdbc:solr://{}:{}?collection={}'.format(host, port, collection)
        self.user = user
        self.password = password

    def query(self, q):
        con_properties = {'user': self.user, 'password': self.password}
        con = jaydebeapi.connect('org.apache.solr.client.solrj.io.sql.DriverImpl',
                                 self.url, con_properties)

        df = pd.read_sql(q, con=con)

        con.close()

        return df, self._make_colmap(df)
    
    def name(self):
        return 'Solr - {}'.format(self._query)



import unittest
import trino
#from common import DB_CREDENTIALS, break_dataset


@unittest.skip("Ignore untill we have environment ready")
class TrinoDS(unittest.TestCase):
    def setUp(self):
        self.HOST = DB_CREDENTIALS['trino']['host']
        self.PORT = DB_CREDENTIALS['trino']['port']
        self.USER = DB_CREDENTIALS['trino']['user']
        self.CATALOG = 'test-catalog'
        self.SCHEMA = 'test-schema'

    def test_trino_ds(self):
        from mindsdb_datasources import TrinoDS

        LIMIT = 100

        trino_ds = TrinoDS(
            host=self.HOST,
            port=self.PORT,
            user=self.USER,
            catalog=self.CATALOG,
            schema=self.SCHEMA,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test.data',
                self.TABLE,
                LIMIT
            )
        )

        trino_ds.df = break_dataset(trino_ds.df)

        assert len(trino_ds) == LIMIT





import os
import unittest
from common import DB_CREDENTIALS


class TestSnowflake(unittest.TestCase):
    def test_snowflake_ds(self):
        print('RUNNING!')
        from mindsdb_datasources import SnowflakeDS

        # Create the datasource
        snowflake_ds = SnowflakeDS(
            query='SELECT * FROM HEALTHCARE_COSTS',
            host=DB_CREDENTIALS['snowflake']['host'],
            user=DB_CREDENTIALS['snowflake']['user'],
            password=DB_CREDENTIALS['snowflake']['password'],
            account=DB_CREDENTIALS['snowflake']['account'],
            warehouse=DB_CREDENTIALS['snowflake']['warehouse'],
            database=DB_CREDENTIALS['snowflake']['database'],
            schema=DB_CREDENTIALS['snowflake']['schema'],
            protocol=DB_CREDENTIALS['snowflake']['protocol'],
            port=DB_CREDENTIALS['snowflake']['port'],
        )

        assert len(snowflake_ds.df) == 1338
        assert len(snowflake_ds.df.columns) == 7



import pandas as pd
import trino

from mindsdb_datasources.datasources.data_source import SQLDataSource


class TrinoDS(SQLDataSource):
    def __init__(self, query, user, password, host='localhost',
                 port=8080, catalog=None, schema='default'):
        super().__init__(query=query)
        self.user = user
        self.password = password
        self.host = host
        self.port = int(port)
        self.catalog = catalog
        self.schema = schema

    def query(self, q):

        con = trino.dbapi.connect(
            self.host,
            self.port,
            self.user,
            catalog=self.catalog, 
            schema=self.schema 
        )

        df = pd.read_sql(q, con=con)
        con.close()

        return df, self._make_colmap(df)

    def name(self):
        return 'Trino - {}'.format(self._query)



from mindsdb_datasources.datasources.data_source import PostgresDS


class TimescaleDS(PostgresDS):
    def __init__(self,query, database, host,
                 port, user, password):
        super.__init__(query, database, host, port, user, password)

    def name(self):
        return 'Timescale - {}'.format(self._query)


import unittest
from common import DB_CREDENTIALS, break_dataset


@unittest.skip("Skip untill we resolve dependency issue")
class TestPhoenix(unittest.TestCase):
    def setUp(self):
        self.URL = DB_CREDENTIALS['phoenix']['url']
        self.AUTHENTICATION = 'BASIC'
        self.USER = DB_CREDENTIALS['phoenix']['user']
        self.PASSWORD = DB_CREDENTIALS['phoenix']['password']
        self.TABLE = 'phoenix_test_table'

    def test_phoenix_ds(self):
        from mindsdb_datasources import PhoenixDS

        LIMIT = 100

        phoenix_ds = PhoenixDS(
            url=self.URL,
            authentication=self.AUTHENTICATION,
            user=self.USER,
            password=self.PASSWORD,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        phoenix_ds.df = break_dataset(phoenix_ds.df)

        assert len(phoenix_ds) == LIMIT



import pandas as pd
import pymssql
from mindsdb_datasources.datasources.data_source import SQLDataSource


class MSSQLDS(SQLDataSource):
    def __init__(self, query, database='master', host='localhost',
               port=1433, user='sa', password=''):
        super().__init__(query=query)
        self.database = database
        self.host = host
        self.port = int(port)
        self.user = user
        self.password = password


    def query(self, q):
        with pymssql.connect(server=self.host, host=self.host, user=self.user, password=self.password, database=self.database, port=self.port) as con:
            df = pd.read_sql(q, con=con)
            
        return df, self._make_colmap(df)

    def name(self):
        return 'Microsoft SQL - {}'.format(self._query)



import setuptools
import sys
import os


def remove_requirements(requirements, name, replace=None):
    new_requirements = []
    for requirement in requirements:
        if requirement.split(' ')[0] != name:
            new_requirements.append(requirement)
        elif replace is not None:
            new_requirements.append(replace)
    return new_requirements

sys_platform = sys.platform

about = {}
with open("mindsdb_datasources/__about__.py") as fp:
    exec(fp.read(), about)

long_description = open('README.md', encoding='utf-8').read()

with open('requirements.txt', 'r') as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

with open('requirements_test.txt', 'r') as req_file:
    test_requirements = [req.strip() for req in req_file.read().splitlines()]

scylla_requirements = []
with open('optional_requirements_scylla.txt', 'r') as fp:
    for line in fp:
        scylla_requirements.append(line.rstrip('\n'))

cassandra_requirements = []
with open('optional_requirements_cassandra.txt', 'r') as fp:
    for line in fp:
        cassandra_requirements.append(line.rstrip('\n'))

extra_data_sources_requirements = []
with open('optional_requirements_extra_data_sources.txt', 'r') as fp:
    for line in fp:
        extra_data_sources_requirements.append(line.rstrip('\n'))

dependency_links = []

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    extras_require = {
        'extra_data_sources': extra_data_sources_requirements,
        'scylla': scylla_requirements,
        'cassandra': cassandra_requirements
    },
    tests_require = test_requirements,
    dependency_links=dependency_links,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    python_requires=">=3.6"
)



import json
import ast
from copy import deepcopy

import pandas as pd
import numpy as np
import moz_sql_parser
from moz_sql_parser.keywords import binary_ops
import traceback


def try_convert_to_dict(val):
    if isinstance(val, dict):
        return val
    if pd.notnull(val):
        try:
            obj = json.loads(val)
            if isinstance(obj, dict):
                return obj
            else:
                raise Exception('Not a json dictionary (could be an int because json.loads is weird)!')
        except Exception as e:
            obj = ast.literal_eval(val)
            if isinstance(obj, dict):
                return obj
            else:
                raise Exception('Expression failed to evaluate to a dictionary')
            return obj
    else:
        return {}


def unnest_df(df):
    original_columns = df.columns
    unnested = 0
    for col in original_columns:
        try:
            json_col = df[col].apply(try_convert_to_dict)
            if np.sum(len(x) for x in json_col) == 0:
                raise Exception('Empty column !')
        except Exception as e:
            continue

        unnested += 1
        unnested_df = pd.json_normalize(json_col)
        unnested_df.columns = [col + '.' + str(subcol) for subcol in unnested_df.columns]
        df = df.drop(columns=[col])

        for unnested_col in unnested_df.columns:
            df[unnested_col] = unnested_df[unnested_col]

    return df, unnested


def unnest(df, col_map):
    df, unnested = unnest_df(df)
    if unnested > 0:
        col_map = {}
        for col in df.columns:
            col_map[col] = col
    return df, col_map

class DataSource:
    def __init__(self, df=None, query=None):
        if type(self) is DataSource and df is None:
            raise Exception('When you\'re instantiating DataSource, you must provide :param df:')

        self._query = query

        if df is not None:
            self._internal_df = df
            self._internal_col_map = self._make_colmap(df)
            self._internal_df, self._internal_col_map = unnest(self._internal_df, self._internal_col_map)
        else:
            self._internal_df = None
            self._internal_col_map = None

        self.data_types = {}
        self.data_subtypes = {}

    def __len__(self):
        return len(self.df)

    def _make_colmap(self, df):
        col_map = {}
        for col in df.columns:
            col_map[col] = col
        return col_map

    def _extract_and_map(self):
        if self._internal_df is None:
            self._internal_df, self._internal_col_map = self.query(self._query)
            self._internal_df, self._internal_col_map = unnest(self._internal_df, self._internal_col_map)

    @property
    def df(self):
        self._extract_and_map()
        return self._internal_df

    @df.setter
    def df(self, df):
        self._internal_df = df

    @property
    def _col_map(self):
        self._extract_and_map()
        return self._internal_col_map

    @_col_map.setter
    def col_map(self, _col_map):
        self._internal_col_map = _col_map

    def _set_df(self, df, col_map):
        self._internal_df = df
        self._col_map = col_map

    def drop_columns(self, column_list):
        """
        Drop columns by original names

        :param column_list: a list of columns that you want to drop
        """
        columns_to_drop = []

        for col in column_list:
            if col not in self._col_map:
                columns_to_drop.append(col)
            else:
                columns_to_drop.append(self._col_map[col])

        self._internal_df.drop(columns=columns_to_drop, inplace=True)

    def query(self, q=None):
        """
        :param q: a query specific to type of datasource
        Datasources must override this method to return pandas.DataFrame
        based on :param q:
        e.g. for MySqlDS :param q: must be a SQL query
             for MongoDS :param q: must be a dictionary

        :return: tuple(pandas.DataFrame, dict)
        """

        # If it's not a subclass of DataSource, then dataframe was provided
        # in the constructor and we just return it
        if type(self) is DataSource:
            return self._internal_df, self._internal_col_map

        # If it is a subclass of DataSource, then this method must be overriden
        else:
            raise NotImplementedError('You must override DataSource.query')

    def _filter_df(self, raw_condition, df):
        """Convert filter conditions to a paticular
        DataFrame instance"""
        col, cond, val = raw_condition
        cond = cond.lower()
        df = df[df[col].notnull()]

        if cond == '>':
            df = df[pd.to_numeric(df[col], errors='coerce') > val]
        if cond == '<':
            df = df[pd.to_numeric(df[col], errors='coerce') < val]
        if cond == 'like':
            df = df[df[col].apply(str).str.contains(str(val).replace("%", ""))]
        if cond == '=':
            df = df[( df[col] == val ) | ( df[col] == str(val) )]
        if cond == '!=':
            df = df[( df[col] != val ) & ( df[col] != str(val) )]

        return df

    def filter(self, where=None, limit=None, get_col_map=False):
        df = self.df
        if where:
            for cond in where:
                df = self._filter_df(cond, df)

        return df.head(limit) if limit else df

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __getattr__(self, attr):
        """
        Map all other functions to the DataFrame
        """
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            return getattr(self._internal_df, attr)

    def __getitem__(self, key):
        """
        Map all other items to the DataFrame
        """
        return self.df.__getitem__(key)

    def __setitem__(self, key, value):
        """
        Support item assignment, mapped to DataFrame
        """
        self.df.__setitem__(key, value)

    def name(self):
        return 'DataFrame'


class SQLDataSource(DataSource):
    def __init__(self, query):
        super().__init__(query=query)

    def filter(self, where=None, limit=None, get_col_map=False):
        try:
            parsed_query = moz_sql_parser.parse(self._query.replace('FORMAT JSON', ''))

            modified_columns = []
            for col, op, value in where or []:
                past_where_clause = parsed_query.get('where', {})

                op = op.lower()
                op_json = binary_ops.get(op, None)

                if op_json is None:
                    print(f"Operator: {op} not found in the sql parser operator list\n Using it anyway.")
                    op_json = op

                if op == 'like':
                    value = '%' + value.strip('%') + '%'
                    if 'clickhouse' in self.name().lower():
                        col = f'toString({col})'
                    elif 'postgres' in self.name().lower():
                        col = f'{col}::text'
                    elif 'mariadb' in self.name().lower() or 'mysql' in self.name().lower() or 'mssql' in self.name().lower():
                        col = f'CAST({col} AS TEXT)'

                modified_columns.append(col)

                where_clause = {op_json: [col, value]}

                if len(past_where_clause) > 0:
                    where_clause = {'and': [where_clause, past_where_clause]}

                parsed_query['where'] = where_clause

            if limit is not None:
                parsed_query['limit'] = limit

            query = moz_sql_parser.format(parsed_query)
            query = query.replace('"', "'")
            query = query.replace("'.'",".")

            for col in modified_columns:
                if f"'{col}'" in query:
                    query = query.replace(f"'{col}'", col)

            df, col_map = self.query(query)
            df, col_map = unnest(df, col_map)
            if get_col_map:
                return df, col_map
            else:
                return df

        except Exception as e:
            print(traceback.format_exc())
            print('Failed to filter using SQL: ', e)
            return super().filter(where=where, limit=limit, get_col_map=get_col_map)

    @property
    def _col_map(self):
        if self._internal_col_map is None:
            _, self._internal_col_map = self.filter(where=[], limit=200, get_col_map=True)
        return self._internal_col_map

    def name(self):
        raise NotImplementedError

    def get_columns(self):
        original_query = self._query.strip(' ;\t\n')
        _result, columns = self.query(f"select * from ({original_query}) as a limit 1")
        columns = list(columns.keys())
        return columns

    def get_row_count(self):
        original_query = self._query.strip(' ;\t\n')
        result, _columns = self.query(f"select count(1) as count from ({original_query}) as a")
        return int(result['count'][0])



import unittest
from common import DB_CREDENTIALS, break_dataset


class TestPostgres(unittest.TestCase):
    def setUp(self):
        self.USER = DB_CREDENTIALS['postgres']['user']
        self.PASSWORD = DB_CREDENTIALS['postgres']['password']
        self.HOST = DB_CREDENTIALS['postgres']['host']
        self.PORT = int(DB_CREDENTIALS['postgres']['port'])
        self.DATABASE = 'postgres'
        self.TABLE = 'home_rentals'

    def test_postgres_ds(self):
        from mindsdb_datasources import PostgresDS

        LIMIT = 100

        postgres_ds = PostgresDS(
            host=self.HOST,
            user=self.USER,
            password=self.PASSWORD,
            database=self.DATABASE,
            port=self.PORT,
            query='SELECT * FROM {}.{} LIMIT {}'.format(
                'test_data',
                self.TABLE,
                LIMIT
            )
        )

        postgres_ds.df = break_dataset(postgres_ds.df)

        assert len(postgres_ds) == LIMIT




import os
import unittest

from mindsdb_datasources import FileDS


class TestFileDS(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.issued_csv = "issued.csv"
        with open(cls.issued_csv, "w") as f:
            f.write("x, y\n")
            for _ in range(20):
                f.write('"With a hard g like gift", "With a hard g like gift"\n')
                f.write('"With a hard ""g,"" like ""gift""", "With a hard ""g,"" like ""gift"""\n')

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.issued_csv)

    def test_file_ds_with_issued_csv(self):
        ds = FileDS(self.issued_csv)
        try:
            ds._handle_source()
        except Exception as e:
            assert False, "issue csv were not handled properly: {}".format(e)


if __name__ == '__main__':
    unittest.main()
